from pathlib import Path
from typing import Any
from urllib.error import URLError

from pytubefix import Stream, StreamQuery, YouTube
from pytubefix.exceptions import PytubeFixError as PytubeError

from youtube_downloader.helpers.util import (
    _error,
    check_ffmpeg,
    complete,
    download,
    download_video_wffmpeg,
    getDefaultTitle,
    metadata,
    progress,
    progress_update,
)

global _ATTEMPTS
_ATTEMPTS = 1


def initialize(url: str, **kwargs: Any) -> tuple[Stream, str]:
    global _ATTEMPTS
    try:
        yt = YouTube(
            url=url,
            client="WEB",
            on_complete_callback=complete,
            on_progress_callback=progress_update,
        )
        stream = get_resolution_upto(yt.streams.filter(progressive=True), **kwargs)
        defaultTitle = getDefaultTitle(yt, subtype=stream.subtype)
        metadata.add_title(url, Path(defaultTitle).stem)

        return stream, defaultTitle
    except URLError:
        if _ATTEMPTS < 4:
            print("\nConnection Error !!! Trying again ... ")
            _ATTEMPTS += 1
            return initialize(url)
        else:
            _error(Exception("Cannot connect to Youtube !!!"))
    except PytubeError as err:
        _error(err)


def get_resolution_upto(streams: StreamQuery, max_res: int = 1080) -> Stream:
    return sorted(
        filter(
            lambda s: (int(s.resolution[:-1]) <= max_res) or (max_res < 0),
            streams.filter(only_video=True),
        ),
        key=lambda s: int(s.resolution[:-1]),
    )[-1]


def initialize_wffmpeg(url: str, **kwargs: Any) -> tuple[Stream, Stream, str]:
    """
    With ffmpeg available, get audio & stream separately.
    return AudioStream, VideoStream, DefaultTitle
    """
    global _ATTEMPTS
    try:
        yt = YouTube(
            url=url,
            client="WEB",
            on_complete_callback=complete,
            on_progress_callback=progress_update,
        )

        audio_stream, video_stream, defaultTitle = _init_ffmpeg(yt, **kwargs)
        metadata.add_title(url, Path(defaultTitle).stem)

        return audio_stream, video_stream, defaultTitle
    except URLError:
        if _ATTEMPTS < 4:
            print("\nConnection Error !!! Trying again ... ")
            _ATTEMPTS += 1
            return initialize_wffmpeg(url)
        else:
            _error(Exception("Cannot connect to Youtube !!!"))
    except PytubeError as err:
        _error(err)


def _init_ffmpeg(yt: YouTube, **kwargs: Any) -> tuple[Stream, Stream, str]:
    audio_stream = yt.streams.get_audio_only()
    video_stream = get_resolution_upto(yt.streams.filter(only_video=True, subtype="mp4"), **kwargs)
    defaultTitle = getDefaultTitle(yt, video_stream.subtype)

    return audio_stream, video_stream, defaultTitle


def get_video(url: str, save_dir: Path, **kwargs: Any) -> None:
    with progress, metadata:
        task_id = progress.custom_add_task(
            title=url,
            description="Downloading",
            start=False,
            total=0,
            completed=0,
        )
        if not check_ffmpeg():
            stream, defaultTitle = initialize(url, **kwargs)
            if not save_dir.joinpath(defaultTitle).exists():
                print(f"+ Downloading resolution {stream.resolution} for {defaultTitle}")
                progress.start_task(task_id)
                progress.update(
                    task_id,
                    description=defaultTitle,
                    total=stream.filesize,
                    completed=0,
                )
                progress.update_mapping(stream.title, task_id)

                download(stream, save_dir, defaultTitle, **kwargs)
        else:
            audio_stream, video_stream, defaultTitle = initialize_wffmpeg(url, **kwargs)
            if not save_dir.joinpath(defaultTitle).exists():
                print(f"+ Downloading resolution {video_stream.resolution} for {defaultTitle}")
                progress.start_task(task_id)
                progress.update(
                    task_id,
                    description=defaultTitle,
                    total=audio_stream.filesize + video_stream.filesize,
                    completed=0,
                )
                progress.update_mapping(audio_stream.title, task_id)
                progress.update_mapping(video_stream.title, task_id)

                download_video_wffmpeg(
                    audio_stream, video_stream, save_dir, defaultTitle, **kwargs
                )
            else:
                progress.remove_task(task_id)


if __name__ == "__main__":
    pass



from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any
from urllib.error import URLError

from pytubefix import Playlist
from pytubefix.exceptions import PytubeFixError as PytubeError

from youtube_downloader.helpers.DownloadAudio import initialize as init_one
from youtube_downloader.helpers.util import (
    NO_WORKER,
    _error,
    getDefaultTitle,
    metadata,
    progress,
    wait,
)
from youtube_downloader.helpers.util import download as download_one

global _ATTEMPTS
_ATTEMPTS = 1


def initialize(url: str) -> Iterable[str]:
    global _ATTEMPTS
    try:
        playlist = Playlist(url, client="WEB")
        metadata.add_title(url, Path(getDefaultTitle(playlist)).stem)
        return playlist.video_urls
    except URLError:
        if _ATTEMPTS < 4:
            print("Connection Error !!! Trying again ... ")
            _ATTEMPTS += 1
            return initialize(url)
        else:
            _error(Exception("Cannot connect to Youtube !!!"))
    except PytubeError as err:
        _error(err)


def download(urls: Iterable[str], save_dir: Path, **kwargs: Any) -> None:
    def run(url: str, **kwargs: Any) -> None:
        task_id = progress.custom_add_task(
            title=url,
            description="Downloading ...",
            total=0,
            completed=0,
        )
        stream, defaultTitle = init_one(url, **kwargs)
        progress.update(task_id, description=stream.title, total=stream.filesize)
        progress.update_mapping(stream.title, task_id)
        download_one(stream, save_dir, defaultTitle, **kwargs)

    with progress:
        with ThreadPoolExecutor(max_workers=NO_WORKER) as pool:
            for url in urls:
                pool.submit(run, url, **kwargs)
                wait(0.5)


def get_audios(url: str, save_dir: Path, **kwargs: Any) -> None:
    with metadata:
        urls = initialize(url)
        download(urls, save_dir, **kwargs)


if __name__ == "__main__":
    pass



import io
from collections.abc import Callable, Iterable
from pathlib import Path
from urllib.parse import urlparse

import click
import pyperclip
import questionary
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import CompleteEvent, Completer, Completion, DeduplicateCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.history import FileHistory, History
from prompt_toolkit.shortcuts import CompleteStyle
from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.markdown import Markdown
from rich.theme import Theme

from youtube_downloader import __version__, scriptDir
from youtube_downloader.helpers import (
    APPDATA,
    YouTubeMetadataCache,
    get_audio,
    get_audios,
    get_video,
    get_video_srt,
    get_videos,
    metadata,
)


def url_validate(url: str) -> bool | str:
    if urlparse(url).netloc.endswith("youtube.com") or (len(url) == 0):
        return True
    else:
        return "Please enter a YouTube URL"


# fmt: off
main_opts: dict[str, Callable[[str, Path], None]] = {
    "1. Download audio only "           : get_audio,
    "2. Download video "                : get_video,
    "3. Download video with caption "   : get_video_srt,
    "4. Download audios from playlist " : get_audios,
    "5. Download videos from playlist " : get_videos,
}
# fmt: on

# fmt: off
res_opts: dict[str, int] = {
    "SD - 480p"      : 480,
    "HD - 720p"      : 720,
    "FullHD - 1080p" : 1080,
    "QHD - 1440p"    : 1440,
    "4K - 2160p"     : 2160,
    "best"           : -1,
}
# fmt: on


class HistoryCompleter(Completer):
    def __init__(self, history: History, metadata: YouTubeMetadataCache) -> None:
        self.history = history
        self.metadata = metadata
        self._merged_data = sorted(
            (
                *history.load_history_strings(),
                *metadata.urls,
            ),
            key=lambda x: metadata.titles.get(x, "unknown"),
        )

    def get_completions(self, document: Document, event: CompleteEvent) -> Iterable[Completion]:
        text = document.text
        for item in self._merged_data:
            if text in item:
                yield Completion(
                    text=item,
                    start_position=-len(text),
                    display_meta=self.metadata.titles.get(item, "unknown"),
                )


url_hist_path = APPDATA / "url_history"
url_hist = FileHistory(url_hist_path)
save_hist_path = APPDATA / "save_history"
save_hist = FileHistory(save_hist_path)

if scriptDir.joinpath("README").exists():
    README_MD = scriptDir.joinpath("README").read_text()
else:
    README_MD = scriptDir.parent.joinpath("README.md").read_text()

USAGE_MARKUP = """
SAMPLE : Provide inputs as per steps below
    [not underline green]STEP 1[/not underline green] : Please enter YouTube URL that you want to download from
      Ex: [bold]? Enter YouTube URL: [yellow]https://youtube.com/?v=example1[/yellow][/bold]

    [not underline green]STEP 2[/not underline green] : Select action
      Ex: [bold]? What do you want to download ? (Use arrow keys)
          [not underline white]1. Download audio only [/not underline white]
        » [not underline reverse white]2. Download video [/not underline reverse white] 
          [not underline white]3. Download video with caption [/not underline white]
          [not underline white]4. Download audios from playlist [/not underline white]
          [not underline white]5. Download videos from playlist [/not underline white]
        [/bold]

    Option 2 selected:
      Ex: [bold]? What do you want to download ? [not underline yellow]2. Download video[/not underline yellow][/bold]

    [not underline green]STEP 3[/not underline green] : Choose save location
      Ex: [bold]? Where do you want to save ? [yellow]~/Videos/[/yellow][/bold]

    [not underline green]STEP 4[/not underline green] : If option 2, 3, or 5 is chosen in Step 2, please select a preferred resolution for video downloading.
      Ex: [bold]? What is your prefered resolution ? (Use arrow keys)
          [white]SD - 480p[/white]
          [white]HD - 720p[/white]
        » [reverse white]FullHD - 1080p[/reverse white]
          [white]QHD - 1440p[/white]
          [white]4K - 2160p[/white]
          [white]best[/white][/bold]

    [not underline green]STEP 5[/not underline green] : If option 3 is chosen in STEP 2, select captions to download here
      Ex: [bold]? Select captions to download (Use arrow keys to move, <space> to select, <a> to toggle, <i> to invert)
            » [reverse white]● en ---- English[/reverse white]
              [white]○ ja ---- Japanese[/white]
              [white]○ ko ---- Korean[/white][/bold]

"""

CMD_HELP_MARKUP = "[italic green]Run without option to start the app[/italic green]"


HelpTheme = Theme(
    styles={
        "repr.switch": "blue",
        "repr.option": "green",
        "repr.number": "bright_blue underline",
        "repr.tag_name": "default",
        "rule.line": "bold yellow",
        "markdown.hr": "bold yellow",
        "markdown.h2": "reverse",
        "markdown.h3": "underline",
    }
)


class RichCommandHighlighter(ReprHighlighter):
    highlights = (
        *ReprHighlighter.highlights,
        r"(^|\W)(?P<switch>\-\w+)(?![a-zA-Z0-9])",
        r"(^|\W)(?P<option>\-\-[\w\-]+)(?![a-zA-Z0-9])",
    )


class RichFormatter(click.HelpFormatter):
    def __init__(
        self,
        indent_increment: int = 2,
        width: int | None = None,
        max_width: int | None = None,
    ) -> None:
        super().__init__(indent_increment, width, max_width)
        self.buffer = io.StringIO()
        self.console = Console(
            file=self.buffer,
            force_terminal=True,
            highlighter=RichCommandHighlighter(),
            theme=HelpTheme,
        )

    def write(self, string: str) -> None:
        self.console.print(string, end="")

    def getvalue(self) -> str:
        return self.buffer.getvalue()

    def write_usage(self, prog: str, args: str = "", prefix: str | None = None) -> None:
        super().write_usage(prog, args, prefix=(prefix or "USAGE : "))


class RichHelpCmd(click.Command):
    def get_help(self, ctx: click.Context) -> str:
        formatter = RichFormatter(width=ctx.terminal_width, max_width=ctx.max_content_width)
        self.format_help(ctx, formatter)
        return formatter.getvalue().rstrip("\n")

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        formatter.write(
            "[bold]youtube-downloader-cli :penguin: CLI APP to download from YouTube[/bold]\n"
        )

        formatter.write_paragraph()
        self.format_usage(ctx, formatter)
        formatter.write(self.help)

        formatter.write_paragraph()
        self.format_options(ctx, formatter)


@click.command(cls=RichHelpCmd, help=CMD_HELP_MARKUP, no_args_is_help=False)
@click.help_option("-h", "--help", is_flag=True, is_eager=True)
@click.option("-v", "--version", is_flag=True, is_eager=True, help="Show version and exit.")
@click.option("-m", "--manual", is_flag=True, is_eager=True, help="Show manual page and exit.")
@click.pass_context
def main(ctx: click.Context, version: bool, manual: bool) -> None:
    if version:
        print(f"youtube-downloader-cli v{__version__}")
        return

    if manual:
        console = Console(
            force_terminal=True,
            highlighter=RichCommandHighlighter(),
            theme=HelpTheme,
        )

        console.print(Markdown(README_MD))
        console.rule()
        print(main.get_help(ctx))
        console.print(USAGE_MARKUP.rstrip())

        return

    if not APPDATA.exists():
        APPDATA.mkdir(parents=True)
    if not url_hist_path.exists():
        url_hist_path.touch()
    if not save_hist_path.exists():
        save_hist_path.touch()

    # Get URL from clipboard if available
    if not pyperclip.is_available():
        txt = pyperclip.paste()
        if url_validate(txt) is not True:
            txt = ""
    else:
        txt = ""

    # Get user inputs (URL, action, save location)
    answers = questionary.form(
        url=questionary.text(
            message="Enter YouTube URL:",
            bottom_toolbar="Use <tab> for history",
            default=txt,
            validate=url_validate,
            enable_history_search=False,
            history=url_hist,
            auto_suggest=AutoSuggestFromHistory(),
            completer=DeduplicateCompleter(HistoryCompleter(url_hist, metadata)),
            complete_style=CompleteStyle.MULTI_COLUMN,
        ),
        opt=questionary.select(message="What do you want to download ?", choices=main_opts),
        loc=questionary.path(
            message="Where do you want to save ?",
            bottom_toolbar="Use <up> & <down> for history",
            default=str(Path.cwd()),
            only_directories=True,
            validate=lambda p: (
                True if Path(p).expanduser().is_dir() else "Please enter path to a directory"
            ),
            enable_history_search=False,
            history=save_hist,
            auto_suggest=AutoSuggestFromHistory(),
        ),
    ).ask()

    # If user cancelled, stop executing
    if not answers:
        return

    save_dir = Path(answers.get("loc")).expanduser().resolve()
    url = answers.get("url")
    opt = answers.get("opt")

    if int(opt[0]) in (2, 3, 5):
        # If downloading video
        res_opt = questionary.select(
            message="What is your prefered resolution ?",
            choices=list(res_opts.keys()),
            default="FullHD - 1080p",
        ).ask()

        if not res_opt:
            return

        max_res = res_opts.get(res_opt)
    else:
        # If downloading audio
        max_res = -1

    main_opts.get(opt)(url, save_dir, max_res=max_res)


if __name__ == "__main__":
    main()



import json
import os
import subprocess
import sys
import unicodedata
from pathlib import Path
from tempfile import TemporaryDirectory
from time import sleep
from typing import Any

from click import ClickException
from pytubefix import Playlist, Stream, YouTube
from rich.console import Console, Group
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    FileSizeColumn,
    GetTimeCallable,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TotalFileSizeColumn,
)
from rich.table import Column
from rich.text import Text

APPDATA = Path("~/.local/share/youtube-downloader-cli/").expanduser()
try:
    NO_WORKER = int(os.environ.get("YTDL_WORKERS", 4))
except ValueError:
    NO_WORKER = 4


class CustomProgress(Progress):
    def __init__(
        self,
        *columns: str | ProgressColumn,
        console: Console | None = None,
        auto_refresh: bool = True,
        refresh_per_second: float = 10,
        speed_estimate_period: float = 30,
        transient: bool = False,
        redirect_stdout: bool = True,
        redirect_stderr: bool = True,
        get_time: GetTimeCallable | None = None,
        disable: bool = False,
        expand: bool = False,
    ) -> None:
        super().__init__(
            *columns,
            console=console,
            auto_refresh=auto_refresh,
            refresh_per_second=refresh_per_second,
            speed_estimate_period=speed_estimate_period,
            transient=transient,
            redirect_stdout=redirect_stdout,
            redirect_stderr=redirect_stderr,
            get_time=get_time,
            disable=disable,
            expand=expand,
        )
        self.task_ids_mapping: dict[str, int] = dict()

    def custom_add_task(
        self,
        title: str,
        description: str,
        start: bool = True,
        total: float | None = 100,
        completed: int = 0,
        visible: bool = True,
        **fields: Any,
    ) -> TaskID:
        task_id = self.add_task(description, start, total, completed, visible, **fields)
        self.task_ids_mapping[title] = task_id
        return task_id

    def update_mapping(self, title: str, task_id: int) -> None:
        self.task_ids_mapping[title] = task_id


progress: CustomProgress = CustomProgress(
    SpinnerColumn(),
    TextColumn(
        "[progress.description]{task.description}",
        table_column=Column(width=30, overflow="ellipsis", no_wrap=True),
    ),
    BarColumn(bar_width=50, style="red", complete_style="green"),
    TextColumn(" - ", style="bar.back"),
    TaskProgressColumn(),
    TextColumn(" - ", style="bar.back"),
    FileSizeColumn(),
    TextColumn("/", style="green"),
    TotalFileSizeColumn(),
    transient=True,
    refresh_per_second=100,
)

progress2: Progress = Progress(
    SpinnerColumn(),
    TextColumn(
        "[progress.description]{task.description}",
        table_column=Column(width=30, overflow="ellipsis", no_wrap=True),
    ),
    BarColumn(bar_width=50, style="red", complete_style="green"),
    TaskProgressColumn(),
    transient=True,
    refresh_per_second=100,
)


def progress_update(stream: Stream, chunk: bytes, bytes_remaining: int) -> None:
    global progress
    task_id = progress.task_ids_mapping.get(stream.title)
    progress.update(task_id, advance=len(chunk))


def complete(stream: Stream, filepath: str) -> None:
    file = Path(filepath)
    task_id = progress.task_ids_mapping.get(stream.title)

    with progress._lock:
        if progress._tasks[task_id].finished:
            if file.stem not in ["ain", "vin"]:
                progress.console.print(
                    f"[green]Successfully downloaded [/green][purple]{file.name} "
                )
            progress.remove_task(task_id)


def download(stream: Stream, save_dir: Path, filename: str, **kwargs: Any) -> None:
    stream.download(filename=filename, output_path=save_dir)


def _error(_exception: Exception) -> None:
    raise ClickException(f"{type(_exception).__name__} : {_exception}")


def getDefaultTitle(y: YouTube | Stream | Playlist, subtype: str = "mp4") -> str:
    """
    Create safe file name by removing special character
    from YouTube video title
    """

    if isinstance(y, Playlist):
        return f"Playlist {y.title}"

    if isinstance(y, YouTube):
        title = (
            y.vid_info.get("microformat", {})
            .get("playerMicroformatRenderer", {})
            .get("title", {})
            .get("simpleText", None)
            or y.title
        ) + f".{subtype}"

    if isinstance(y, Stream):
        title = y.default_filename

    return title.translate(
        {
            ord(x): ""
            for x in title
            if unicodedata.category(x)[0] not in "LN" and x not in "_-()[]! ."
        }
    )


def check_ffmpeg() -> bool:
    """
    Check if ffmpeg is available.
    return True if ffmpeg is present, False otherwise
    """

    try:
        subprocess.check_call(
            ["ffmpeg", "-version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        progress.console.print("[yellow]ffmpeg is not available", file=sys.stderr)
        return False


def ffmpeg_merge(audio: Path, video: Path, out: Path) -> bool:
    try:
        # fmt: off
        subprocess.check_output(
            [
                "ffmpeg", "-y",
                "-i", str(audio.resolve()),
                "-i", str(video.resolve()),
                "-c", "copy",
                "-loglevel", "warning",
                str(out.resolve()),
            ],
            stderr=subprocess.STDOUT,
        )
        # fmt: on
        return True
    except subprocess.CalledProcessError as e:
        panel = Panel(
            Group(
                Text.assemble(("failed to merge ", "default"), (out.name, "purple")),
                Panel(
                    e.output.decode().strip(),
                    title="[red]error",
                    title_align="left",
                    border_style="red",
                ),
            ),
            title="[red]error using ffmpeg",
            title_align="left",
            border_style="red",
        )
        progress.console.print(panel)
        return False


def download_video_wffmpeg(
    audio_stream: Stream, video_stream: Stream, save_dir: Path, filename: str, **kwargs: Any
) -> None:
    with TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
        audio_file = Path(tmpdir) / f"ain.{audio_stream.subtype}"
        video_file = Path(tmpdir) / f"vin.{video_stream.subtype}"

        download(audio_stream, Path(tmpdir), audio_file.name, **kwargs)
        download(video_stream, Path(tmpdir), video_file.name, **kwargs)

        if ffmpeg_merge(audio_file, video_file, save_dir / filename):
            progress.console.print(f"[green]Successfully downloaded [/green][purple]{filename}")


def wait(sec: float) -> None:
    sleep(sec)


class YouTubeMetadataCache:
    def __init__(self, p: Path) -> None:
        if not p.exists():
            if not p.parent.exists():
                p.parent.mkdir(parents=True)
            p.touch()
        self._filepath = p
        self._loaded = False

    @property
    def urls(self) -> list[str]:
        if (not self._loaded) or (not hasattr(self, "_urls")):
            self._load()
            self._urls = [url for url in self._json.keys()]
        return self._urls

    @property
    def titles(self) -> dict:
        if (not self._loaded) or (not hasattr(self, "_titles")):
            self._load()
            self._titles = {url: self._json.get(url, {}).get("title", "") for url in self._json}

        return self._titles

    def _load(self) -> None:
        if not self._loaded:
            self._cache = self._filepath.read_bytes()
            if self._cache:
                self._json = json.loads(self._cache)
            else:
                self._json = {}
            self._loaded = True

    def add_title(self, url: str, title: str) -> None:
        if not self._loaded:
            self._load()
        self._json[url] = {}
        self._json[url]["title"] = title
        if not hasattr(self, "_titles"):
            self._titles = {}
        self._titles[url] = title

    def _store(self) -> None:
        if not self._loaded:
            return

        self._filepath.write_text(json.dumps(self._json))

    def __enter__(self):  # noqa: ANN204
        if not self._loaded:
            self._load()
        return self

    def __exit__(self, ext, val, tb) -> None:  # noqa: ANN001
        self._store()


metadata = YouTubeMetadataCache(APPDATA / "metadata")
